﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Web;
using System.Web.Mvc;
using System.Web.Routing;
using MbUnit.Framework;
using Moq;

namespace RouteTesting.Tests.Routes
{
	/* ====== RouteFixtureBase class ====== */

	/// <summary>
	///  This is the base class for a test fixture used to unit test routes
	///  in ASP.NET MVC.
	/// </summary>

	public class RouteFixtureBase
	{
		private RouteCollection routes;


		/* ====== TestFixtureSetUp method ====== */

		/// <summary>
		///  Initialises the routes ready for testing.
		/// </summary>

		[TestFixtureSetUp]
		public void TestFixtureSetUp()
		{
			routes = new RouteCollection();
			RouteTesting.MvcApplication.RegisterRoutes(routes);
		}


		/* ======= PropertyValue nested class ====== */

		/// <summary>
		///  Encapsulates the name and value of a property of an object.
		/// </summary>

		private sealed class PropertyValue
		{
			public string Name { get; set; }
			public object Value { get; set; }
		}


		/* ====== GetProperties method ====== */

		/// <summary>
		///  Reads in all the properties of an object (usually an anonymous class).
		///  Based on the technique outlined by Eilon Lipton at
		///  http://weblogs.asp.net/leftslipper/archive/2007/09/24/using-c-3-0-anonymous-types-as-dictionaries.aspx
		/// </summary>
		/// <param name="o">The object whose properties are being examined.</param>
		/// <returns>An iterator over the properties defined on the object.</returns>

		private static IEnumerable<PropertyValue> GetProperties(object o)
		{
			if (o != null) {
				PropertyDescriptorCollection props
				  = TypeDescriptor.GetProperties(o);
				foreach (PropertyDescriptor prop in props) {
					object val = prop.GetValue(o);
					if (val != null) {
						yield return new PropertyValue { Name = prop.Name, Value = val };
					}
				}
			}
		}


		/* ====== FindUrlToRoute method ====== */

		/// <summary>
		///  Finds the URL for a route given the starting page, and the route data for
		///  the target page. All defined routes will be searched for a match.
		/// </summary>
		/// <param name="currentPage">
		///  The name of the starting page from which the URL is requested.
		/// </param>
		/// <param name="routeData">
		///  An object encapsulating the data used to construct the route.
		/// </param>
		/// <returns>
		///  The URL to the route. If no matching route is found, returns null.
		/// </returns>

		private string FindUrlToRoute(string currentPage, object routeData)
		{
			var mockContext = new Mock<HttpContextBase>();
			var mockRequest = new Mock<HttpRequestBase>();
			var mockResponse = new Mock<HttpResponseBase>();

			mockContext.Expect(c => c.Request).Returns(mockRequest.Object);
			mockContext.Expect(c => c.Response).Returns(mockResponse.Object);
			mockRequest.Expect(c => c.AppRelativeCurrentExecutionFilePath)
				.Returns(currentPage);
			mockResponse.Expect(c => c.ApplyAppPathModifier(It.IsAny<string>()))
				.Returns((string s) => s);

			var route = routes.GetRouteData(mockContext.Object);

			var requestContext
			  = new RequestContext(mockContext.Object, route);
			var url = new UrlHelper(requestContext);
			var dict = new RouteValueDictionary();

			foreach (PropertyValue property in GetProperties(routeData)) {
				dict[property.Name] = property.Value;
			}

			var path = routes.GetVirtualPath(requestContext, dict);
			return (path != null ? path.VirtualPath : null);
		}


		/* ====== AssertRouteUrl method ====== */

		/// <summary>
		///  Asserts that the URL for a given route should be as expected.
		/// </summary>
		/// <param name="currentPage">
		///  The name of the starting page from which the URL is requested.
		/// </param>
		/// <param name="expectedUrl">
		///  The URL that we expect to be returned.
		/// </param>
		/// <param name="routeData">
		///  An object encapsulating the data used to construct the route.
		/// </param>

		protected void AssertRouteUrl(string currentPage, string expectedUrl, object routeData)
		{
			var str = FindUrlToRoute(currentPage, routeData);
			Assert.AreEqual(expectedUrl, str, "URL was wrong!");
		}


		#region /* ====== Route data test ====== */

		/*
		 * Based, with modifications, on 
		 * http://haacked.com/archive/2007/12/17/testing-routes-in-asp.net-mvc.aspx
		 */


		/* ====== NotPresent ====== */

		/// <summary>
		///  A special marker object, when it is present in the expectations, it indicates
		///  that this value should NOT be present in the route data.
		/// </summary>

		public static readonly object NotPresent = new object();


		/* ====== AssertRoute ====== */

		/// <summary>
		///  Tests the route collection to ensure that the data returned by it is correct.
		/// </summary>
		/// <param name="url">
		///  The URL being tested. This will be app relative, so it should begin with a tilde.
		/// </param>
		/// <param name="expectations">
		///  An object encapsulating the expectations on the data to be returned.
		/// </param>

		protected void AssertRoute(string url, object expectations)
		{
			var httpContextMock = new Mock<HttpContextBase>();
			var requestMock = new Mock<HttpRequestBase>();
			httpContextMock.Expect(c => c.Request).Returns(requestMock.Object);
			requestMock.Expect(r => r.AppRelativeCurrentExecutionFilePath)
				 .Returns(url);

			RouteData routeData = routes.GetRouteData(httpContextMock.Object);
			Assert.IsNotNull(routeData, "Should have found the route");

			foreach (PropertyValue property in GetProperties(expectations)) {
				if (object.ReferenceEquals(property.Value, NotPresent)) {
					Assert.IsFalse(
						routeData.Values.ContainsKey(property.Name),
						String.Format("Route data unexpectedly contains {0}={1}",
							property.Name,
							routeData.Values[property.Name]
						)
					);
				}
				else {
					Assert.IsTrue(
						string.Equals(
							property.Value.ToString(),
							routeData.Values[property.Name].ToString(),
							StringComparison.OrdinalIgnoreCase
						),
						string.Format(
							"Expected '{0}', not '{1}' for '{2}'.",
							property.Value, 
							routeData.Values[property.Name], 
							property.Name
						)
					);
				}
			}
		}

		#endregion
	}
}